// Nero MP4 codec Cool Edit filter
// Copyright (C) 2004 Janne Hyvrinen
// Based on NEncode code, Copyright (C) 2003 DPeshev
//
// Changes:
//  0.0.2 (2004-02-08): More accurate decoding length handling
//  0.0.1 (2004-02-08): Initial version

#include <afxwin.h>     // MFC core and standard components
#include <afxext.h>     // MFC extensions
#include <afxdtctl.h>   // MFC support for Internet Explorer 4 Common Controls
#include "neroapiglue.h"
#include "AudioBase.h"
#include <windows.h>
#include <stdlib.h>
#include <direct.h>
#include <math.h>
#include <commctrl.h>
#include "res/filters.h"

#define BPS     16

#define CN_ERRORMESSAGES        // MessageBox on errors
#define CN_MAX_CHANNELS     2
#define CHUNKSIZE   2048 * CN_MAX_CHANNELS * (BPS/8)

void infobox ( const char *msg )
{
# ifdef CN_ERRORMESSAGES
    MessageBox ( 0, msg, "Nero MP4", 0 );
# endif
}

static required_sample_rate = 0;

// -------------------------------------

class CAudioPluginMgr : public IAudioPluginMgr, public IProfile {
public:
    CAudioPluginMgr() : m_lRef(0) {};
    virtual ~CAudioPluginMgr() {}
public:
    // IUnknown
    STDMETHODIMP QueryInterface(REFIID riid, void __RPC_FAR* __RPC_FAR* ppObj);
    STDMETHODIMP_(ULONG) AddRef();
    STDMETHODIMP_(ULONG) Release();
    // IAudioPluginMgr
    virtual bool SetPluginDirectory(const char *szPluginDir);
    virtual bool LoadPlugins();
    virtual bool CanUnloadPlugins();
    virtual bool UnloadPlugins();
    virtual bool SetConvertor(IConvertorFactory *pCF, IStatus **ppStatus = NULL);
    virtual bool GetConvertor(IConvertorFactory **ppConvFactory);
    virtual bool GetUnsupportedComponents(IComponentEnum **ppEnum);
    virtual bool SendEvent(EEvent event, IUnknown *pSource);
    virtual bool SubscribeToEvents(IEventReceiver *pReceiver);
    virtual bool UnsubscribeToEvents(IEventReceiver *pReceiver);
    virtual bool OpenURLAudioFile(  const char          *szURL,
                        const SWavFormat    *pFormat,
                        IUnknown            **ppSrc,
                        int                 iTries,
                        EAuxFlags           flagsInclude,
                        EAuxFlags           flagsExclude,
                        IStatus             **ppStatus);
    virtual bool CreateURLAudioTarget(  IURLAudioTargetFactory  *pFactory,
                            IUnknown                **pTgt,
                            const SWavFormat        &formatSrc,
                            IStatus                 **ppStatus  = NULL);
    virtual bool AttachAuxObject(REFCLSID clsID, IUnknown *pUnknown);
    virtual bool DetachAuxObject(REFCLSID clsID);
    virtual bool GetAuxObject(REFCLSID clsID, IUnknown **ppUnknown);
    virtual bool TranslateError(IStatus *pInStatus, IStatus **ppOutStatus);
    // IProfile
    virtual bool SetVendor(const char *szVendor, IStatus **ppStatus = NULL);
    virtual const char *GetVendor();
    virtual bool SetProduct(const char *szProduct, IStatus **ppStatus = NULL);
    virtual const char * GetProduct();
    // Data
protected:
    long    m_lRef;
};

//IAudioPluginMgr
bool CAudioPluginMgr::SetPluginDirectory(const char *szPluginDir) {return true;};
bool CAudioPluginMgr::LoadPlugins() {return true;};
bool CAudioPluginMgr::CanUnloadPlugins() {return true;};
bool CAudioPluginMgr::UnloadPlugins() {return true;};
bool CAudioPluginMgr::SetConvertor(IConvertorFactory *pCF, IStatus **ppStatus) {return true;};
bool CAudioPluginMgr::GetConvertor(IConvertorFactory **ppConvFactory) {return true;};
bool CAudioPluginMgr::GetUnsupportedComponents(IComponentEnum **ppEnum) {return true;};
bool CAudioPluginMgr::SendEvent(EEvent event, IUnknown *pSource) {return true;};
bool CAudioPluginMgr::SubscribeToEvents(IEventReceiver *pReceiver) {return true;};
bool CAudioPluginMgr::UnsubscribeToEvents(IEventReceiver *pReceiver) {return true;};
bool CAudioPluginMgr::OpenURLAudioFile( const char          *szURL,
                        const SWavFormat    *pFormat,
                        IUnknown            **ppSrc,
                        int                 iTries,
                        EAuxFlags           flagsInclude,
                        EAuxFlags           flagsExclude,
                        IStatus             **ppStatus) {return true;};
bool CAudioPluginMgr::CreateURLAudioTarget( IURLAudioTargetFactory  *pFactory,
                            IUnknown                **pTgt,
                            const SWavFormat        &formatSrc,
                            IStatus                 **ppStatus) {return true;};
bool CAudioPluginMgr::AttachAuxObject(REFCLSID clsID, IUnknown *pUnknown) {return true;};
bool CAudioPluginMgr::DetachAuxObject(REFCLSID clsID) {return true;};
bool CAudioPluginMgr::GetAuxObject(REFCLSID clsID, IUnknown **ppUnknown) {return true;};
bool CAudioPluginMgr::TranslateError(IStatus *pInStatus, IStatus **ppOutStatus) {return true;};
// IProfile
bool CAudioPluginMgr::SetVendor(const char *szVendor, IStatus **ppStatus) {return true;};
//const char *CAudioPluginMgr::GetVendor() { return "Smeshka";};
const char *CAudioPluginMgr::GetVendor() { return "Ahead";};
bool CAudioPluginMgr::SetProduct(const char *szProduct, IStatus **ppStatus) {return true;};
//const char * CAudioPluginMgr::GetProduct() { return "Nencode"; };
const char * CAudioPluginMgr::GetProduct() { return "Nero - Burning Rom"; };
// QueryInterface returns either IUnknown, IAudioPluginMgr or IProfile
STDMETHODIMP CAudioPluginMgr::QueryInterface(REFIID riid, void __RPC_FAR *__RPC_FAR *ppObj) {
  // ppObj must not be NULL
  if(NULL == ppObj) {
    //assert(FALSE);
    infobox ( "QueryInterface received NULL pointer" );
    return E_INVALIDARG;
  }
  *ppObj = NULL;
  if(IsEqualIID(riid, IID_IUnknown)) *ppObj = static_cast<IUnknown*>(static_cast<IAudioPluginMgr*>(this));
  else if(IsEqualIID(riid, IID_IAudioPluginMgr)) *ppObj = static_cast<IAudioPluginMgr*>(this);
  else if(IsEqualIID(riid, IID_IProfile)) *ppObj = static_cast<IProfile*>(this);
  if(*ppObj) AddRef();
  if (NULL != *ppObj) return S_OK;
  else return E_NOINTERFACE;
}

// Increase the reference counter and return the number of references
STDMETHODIMP_(ULONG) CAudioPluginMgr::AddRef() {
  InterlockedIncrement(&m_lRef);
  return m_lRef;
}

// Decrease the reference counter and return the number of references
STDMETHODIMP_(ULONG) CAudioPluginMgr::Release() {
  InterlockedDecrement(&m_lRef);
  return m_lRef;
}

typedef bool (* LPNERO_PLUGIN_GetPrimaryAudioObject)(IAudioComponent**);
typedef bool (* LPNERO_PLUGIN_ReadyToFinish)();

// -------------------------------------

static int get_plugin_dir ( char plugin_dir[_MAX_PATH], char cwd[_MAX_PATH] )
{
    if ( plugin_dir[0] != '\0' ) return 1;

    const static struct {
        HKEY key;
        LPCTSTR subkey;
        LPCTSTR value;
    } locations[] = {
        { HKEY_LOCAL_MACHINE, "Software\\Ahead\\Shared", "AudioPlugIn" },
        { HKEY_CURRENT_USER, "Software\\Ahead\\NeroMIX\\Config", "PluginMgrPath" },
    };

    char tmp[1024];
    if ( !_getcwd (tmp, 1024) ) return 0;
    strncpy ( plugin_dir, tmp, _MAX_PATH-1 );
    plugin_dir[_MAX_PATH-1] = '\0';

    char temp[4096];
    HKEY key;
    bool found = false;

    for ( int pos = 0; pos < sizeof(locations)/sizeof(*locations); pos++ ) {
        if ( RegOpenKeyEx (locations[pos].key, locations[pos].subkey, 0, KEY_READ, &key) == ERROR_SUCCESS ) {
            DWORD size  = sizeof(temp);
            if ( RegQueryValueEx (key, locations[pos].value, NULL, NULL, (LPBYTE)temp, &size) == ERROR_SUCCESS ) found = true;
            RegCloseKey ( key );
            if ( found ) break;
        }
    }

    if ( !found ) return 0;

    strncpy ( plugin_dir, temp, _MAX_PATH-1 );
    plugin_dir[_MAX_PATH-1] = '\0';

    {
        char *p = (char *)plugin_dir + strlen(plugin_dir) - 1;
        while ( p >= plugin_dir && *p == '\\' ) {
            *p-- = '\0';
        }
    }

    if ( plugin_dir[0] == '\0' ) return 0;

    if ( _chdir (plugin_dir) != 0 ) {
        plugin_dir[0] = '\0';
        return 0;
    }

    return 1;
}


// -------------------------------------
// Encoder
// -------------------------------------

class nero_enc {
private:
    void nullify()
    {
        hDLL = 0;
        GetPrimaryAudioObject = 0;
        ReadyToFinish = 0;
        pIA = 0;

        pIC = 0;
        pICU = 0;
        pIAC = 0;
        pIEE = 0;
        pDest = 0;
        pIU = 0;

        pIAURW = 0;
        pIW = 0;
        pIURLH = 0;
        pIProcess = 0;

        initialized = false;
    }

public:
    nero_enc()
    {
        nullify();
        cwd[0] = '\0';
        path[0] = '\0';
        plugin_dir[0] = '\0';
    }

    ~nero_enc()
    {
        close_plugin();

        if ( cwd[0] ) _chdir ( cwd );

        pluginMgr.Release();
    }

    int config()
    {
        if ( hDLL ) close_plugin();

        if ( !get_plugin_dir(plugin_dir, cwd) ) {
            infobox ( "Couldn't find Nero plugin directory" );
            return 0;
        }

        if ( !init_plugin() ) {
            infobox ( "Nero encoder plugin initialization failed" );
            close_plugin();
            return 0;
        }

        int ret = 1;

        if ( pIC ) {
            int nRetCode = pIC->Configure();
            if ( !nRetCode ) ret = 0;
        } else {
            ret = 0;
        }

        close_plugin();

        return ret;
    }

    int init ( const char *out_path, unsigned int srate, unsigned int nch, unsigned int bps, unsigned int *required_srate=0 )
    {
        if ( hDLL ) return 0;

        strncpy ( path, out_path, sizeof(path)-1 );
        path[sizeof(path)-1] = '\0';

        format.m_iSamplesPerSecond = srate;
        format.m_iBitsPerSample = bps;
        format.m_iChannels = nch;

        if ( !get_plugin_dir(plugin_dir, cwd) ) {
            infobox ( "Couldn't find Nero plugin directory" );
            return 0;
        }

        if ( !init_plugin() ) {
            infobox ( "Nero encoder plugin initialization failed" );
            close_plugin();
            return 0;
        }

        if ( !init_encoder() ) {
            infobox ( "Nero encoder plugin initialization failed" );
            close_plugin();
            return 0;
        }

        if ( required_srate ) *required_srate = format_in.m_iSamplesPerSecond;

        return 1;
    }

    int encode ( const void *buffer, unsigned int samples, unsigned int bps, unsigned int nch )
    {
        if ( !initialized ) return 0;

        /*
        if ( (format.m_iSamplesPerSecond != format_in.m_iSamplesPerSecond) && !p_resampler_dsp ) {
            p_resampler_dsp = resampler::create ( format_in.m_iSamplesPerSecond, 0 );
            if ( !p_resampler_dsp ) {
                console::error ( "Failed to initialize resampler" );
                return 0;
            }
        }
        */

        enum EAudioRawState state;
        int nRetCode = pIAURW->RawWrite ( (unsigned char *)buffer, samples * nch * (bps/8), state, &pIStatus );
        if ( !(nRetCode && !pIStatus) ) {
            char p[256];
            const char *pp = pIStatus->GetDescription();
            if ( !pp || (pp && !*pp) ) {
                sprintf ( p, "Plugin returned with error code %i", nRetCode );
            } else {
                strncpy ( p, pp, sizeof(p)-1 );
                p[sizeof(p)-1] = '\0';
            }
            infobox ( p );
            return 0;
        }

        return 1;
    }

private:
    void close_plugin()
    {
        if ( hDLL ) {
            int nRetCode;

            if ( pIProcess ) {
                nRetCode = pIProcess->End ( &pIStatus );
            }

            if ( initialized ) {
                nRetCode = pIAC->Done();
                ReadyToFinish();
            }

            if ( pIProcess ) pIProcess->Release();
            if ( pIURLH ) pIURLH->Release();
            if ( pIW ) pIW->Release();
            if ( pIAURW ) pIAURW->Release();

            if ( pIEE ) pIEE->Release();
            if ( pIC ) pIC->Release();
            if ( pDest ) pDest->Release();
            if ( pIAC ) pIAC->Release();
            if ( pICU ) pICU->Release();
            if ( pIA ) pIA->Release();
            if ( pIStatus ) pIStatus->Release();

            FreeLibrary ( hDLL );
            hDLL = 0;

            nullify();
        }
    }

    int init_plugin()
    {
        initialized = false;

        hDLL = LoadLibraryA ( "aac.dll" );
        if ( !hDLL ) {
            infobox ( "Failed to open plugin 'aac.dll'" );
            return 0;
        }

        GetPrimaryAudioObject = (LPNERO_PLUGIN_GetPrimaryAudioObject)GetProcAddress ( hDLL, "NERO_PLUGIN_GetPrimaryAudioObject" );
        if ( !GetPrimaryAudioObject ) return 0;

        ReadyToFinish = (LPNERO_PLUGIN_ReadyToFinish)GetProcAddress ( hDLL, "NERO_PLUGIN_ReadyToFinish" );
        if ( !ReadyToFinish ) return 0;

        int nRetCode = GetPrimaryAudioObject ( &pIA );
        if ( !pIA || !(nRetCode && pIA->GetType() == ACT_ComponentEnumerator) ) return 0;

        nRetCode = pIA->QueryInterface ( IID_IComponentEnum, (void **)&pICU );
        if ( !(!nRetCode && pICU && pICU->GetCount()>1) ) return 0;

        nRetCode = pICU->GetComponent ( 1, &pIAC );
        if ( !(nRetCode && pIAC && pIAC->GetType() == ACT_AudioURLTargetFactory) ) return 0;

        nRetCode = pIAC->QueryInterface ( IID_IURLAudioTargetFactory, (void**)&pDest );
        if ( !(!nRetCode && pDest) ) return 0;

        nRetCode = pIAC->QueryInterface ( IID_IConfigurable, (void**)&pIC );
        if ( !(!nRetCode && pIC) ) pIC = 0; // return 0;

        nRetCode = pIAC->QueryInterface ( IID_IExtEnum, (void**)&pIEE );
        if ( !(!nRetCode && pIEE) ) return 0;

        nRetCode = pIAC->Init ( static_cast<IAudioPluginMgr*>(&pluginMgr), &pIStatus );
        if ( !(nRetCode && !pIStatus) ) return 0;

        initialized = true;

        /*
        plugin_name.set_string_ansi ( pIAC->GetName() );

        for ( int i = 0; i < pIEE->GetCount(); i++ ) {
            const char *ext = pIEE->GetExt ( i );

            if ( ext ) {
                string8 *e = new string8;
                string8 tmp;
                tmp.set_string_ansi ( ext );
                e->set_string_lower ( (const char *)tmp );
                plugin_ext.add_item ( e );
            }
        }
        */

        return 1;
    }

    int init_encoder()
    {
        if ( !initialized ) return 0;

        int nRetCode = pDest->CreateURLAudioTarget ( &pIU, format, &pIStatus );
        if ( !(nRetCode && !pIStatus) ) return 0;

        nRetCode = pIU->QueryInterface ( IID_IAudioRawWriter, (void**)&pIAURW );
        if ( !(!nRetCode && pIAURW) ) return 0;

        format_in = pIAURW->GetRawFormat();

        nRetCode = pIU->QueryInterface ( IID_IInfoWriter, (void**)&pIW );
        if ( !(!nRetCode && pIW) ) pIW = 0; // return 0;

        nRetCode = pIU->QueryInterface ( IID_IURLHolder, (void**)&pIURLH );
        if ( !(!nRetCode && pIURLH) ) return 0;

        nRetCode = pIURLH->SetURL ( (const char *)path, &pIStatus );
        if ( !(nRetCode && !pIStatus) ) return 0;

        nRetCode = pIU->QueryInterface ( IID_IProcess, (void**)&pIProcess );
        if ( !(!nRetCode && pIProcess) ) return 0;

        nRetCode = pIProcess->Start ( &pIStatus );
        if ( !(nRetCode && !pIStatus) ) return 0;

        return 1;
    }

private:
    char path[_MAX_PATH];
    char cwd[_MAX_PATH];
    char plugin_dir[_MAX_PATH];
    bool initialized;

    SWavFormat format;
    SWavFormat format_in;

    CAudioPluginMgr pluginMgr;
    IStatus *pIStatus;

    HINSTANCE hDLL;
    LPNERO_PLUGIN_GetPrimaryAudioObject GetPrimaryAudioObject;
    LPNERO_PLUGIN_ReadyToFinish ReadyToFinish;
    IAudioComponent *pIA;

    IConfigurable *pIC;
    IComponentEnum *pICU;
    IAudioComponent *pIAC;
    IExtEnum *pIEE;
    IURLAudioTargetFactory *pDest;
    IUnknown *pIU;

    IAudioRawWriter *pIAURW;
    IInfoWriter *pIW;
    IURLHolder *pIURLH;
    IProcess *pIProcess;
};

// -------------------------------------
// Decoder
// -------------------------------------

class nero_dec {
private:
    void nullify()
    {
        hDLL = 0;
        GetPrimaryAudioObject = 0;
        ReadyToFinish = 0;
        pIA = 0;

        pICU = 0;
        pIAC = 0;
        pIEE = 0;
        pSrc = 0;
        pIU = 0;
        pIEAI = 0;
        pIS = 0;
        pIS = 0;
        pIAS = 0;

        pIAURR = 0;
        pIProcess = 0;
    }

public:
    nero_dec()
    {
        nullify();
        initialized = false;
        //buffer = 0;
        cwd[0] = '\0';
        path[0] = '\0';
        plugin_dir[0] = '\0';
    }

    ~nero_dec()
    {
        close_plugin();

        if ( cwd[0] ) _chdir ( cwd );

        pluginMgr.Release();
    }

    int init ( const char *in_path, unsigned int *srate=0, unsigned int *nch=0, unsigned int *bps=0, double *seconds=0, double *bitrate=0 )
    {
        if ( hDLL ) return 0;

        strncpy ( path, in_path, sizeof(path)-1 );
        path[sizeof(path)-1] = '\0';

        if ( !get_plugin_dir(plugin_dir, cwd) ) {
            infobox ( "Couldn't find Nero plugin directory" );
            return 0;
        }

        if ( !init_plugin() ) {
            infobox ( "Nero decoder plugin initialization failed" );
            close_plugin();
            return 0;
        }

        if ( !init_decoder() ) {
            infobox ( "Nero decoder plugin initialization failed" );
            close_plugin();
            return 0;
        }

        /*
        buffer = (unsigned char *)malloc ( 2048*format.m_iChannels*(BPS/8) );
        if ( !buffer ) {
            close_plugin();
            return 0;
        }
        */

        if ( srate ) *srate = format.m_iSamplesPerSecond;
        if ( bps ) *bps = format.m_iBitsPerSample;
        if ( nch ) *nch = format.m_iChannels;
        if ( seconds ) *seconds = (double)_duration / 1000.0;
        if ( bitrate ) *bitrate = (double)_bitrate / 1000.0;

        return 1;
    }

    /*
    bool seek ( double pos )
    {
        if ( !pIS ) return false;
        return pIS->Seek ( (ULONGLONG)(pos * (double)format.m_iSamplesPerSecond), &pIStatus );
    }
    */

    int decode ( void *out_buffer, unsigned int *out_size, unsigned int *srate, unsigned int *bps, unsigned int *nch )
    {
        if ( !initialized ) return 0;

        enum EAudioRawState state;
        int blockSize = 2048 * format.m_iChannels * (BPS/8);
        int blockSizeRead = 0;

        int nRetCode = pIAURR->RawRead ( (unsigned char *)out_buffer, blockSize, &blockSizeRead, state, &pIStatus );
        if ( (state == ERS_EndOfFile) || (!blockSizeRead) ) return 0;
        if ( !(nRetCode && !pIStatus) ) {
            char p[256];
            const char *pp = pIStatus->GetDescription();
            if ( !pp || (pp && !*pp) ) {
                sprintf ( p, "Plugin returned with error code %i", nRetCode );
            } else {
                strncpy ( p, pp, sizeof(p)-1 );
                p[sizeof(p)-1] = '\0';
            }
            infobox ( p );
            return 0;
        }

        //*out_buffer = (void *)buffer;
        *out_size = blockSizeRead;

        *srate = format.m_iSamplesPerSecond;
        *bps = format.m_iBitsPerSample;
        *nch = format.m_iChannels;

        return 1;
    }

private:
    void close_plugin()
    {
        if ( hDLL ) {
            int nRetCode;

            if ( pIProcess ) {
                nRetCode = pIProcess->End ( &pIStatus );
            }

            if ( initialized ) {
                nRetCode = pIAC->Done();
                ReadyToFinish();
            }

            if ( pIAS ) pIAS->Release();
            if ( pIR ) pIR->Release();
            if ( pIS ) pIS->Release();
            if ( pIEAI ) pIEAI->Release();
            if ( pIProcess ) pIProcess->Release();
            if ( pIAURR ) pIAURR->Release();
            if ( pIU ) pIU->Release();

            if ( pIEE ) pIEE->Release();
            if ( pSrc ) pSrc->Release();
            if ( pIAC ) pIAC->Release();
            if ( pICU ) pICU->Release();
            if ( pIA ) pIA->Release();
            if ( pIStatus ) pIStatus->Release();

            FreeLibrary ( hDLL );
            hDLL = 0;

            nullify();
        }

        /*
        if ( buffer ) {
            free ( buffer );
            buffer = 0;
        }
        */
    }

    int init_plugin()
    {
        initialized = false;

        hDLL = LoadLibraryA ( "aac.dll" );
        if ( !hDLL ) {
            infobox ( "Failed to open plugin 'aac.dll'" );
            return 0;
        }

        GetPrimaryAudioObject = (LPNERO_PLUGIN_GetPrimaryAudioObject)GetProcAddress ( hDLL, "NERO_PLUGIN_GetPrimaryAudioObject" );
        if ( !GetPrimaryAudioObject ) return 0;

        ReadyToFinish = (LPNERO_PLUGIN_ReadyToFinish)GetProcAddress ( hDLL, "NERO_PLUGIN_ReadyToFinish" );
        if ( !ReadyToFinish ) return 0;

        int nRetCode = GetPrimaryAudioObject ( &pIA );
        if ( !(nRetCode && pIA->GetType() == ACT_ComponentEnumerator) ) return 0;

        nRetCode = pIA->QueryInterface ( IID_IComponentEnum, (void **)&pICU );
        if ( !(!nRetCode && pICU && pICU->GetCount()>0) ) return 0;

        nRetCode = pICU->GetComponent ( 0, &pIAC );
        if ( !(nRetCode && pIAC && pIAC->GetType() == ACT_AudioURLSourceFactory) ) return 0;

        nRetCode = pIAC->QueryInterface ( IID_IURLAudioSourceFactory, (void**)&pSrc );
        if ( !(!nRetCode && pSrc) ) return 0;

        nRetCode = pIAC->QueryInterface ( IID_IExtEnum, (void**)&pIEE );
        if ( !(!nRetCode && pIEE) ) return 0;

        nRetCode = pIAC->Init ( static_cast<IAudioPluginMgr*>(&pluginMgr), &pIStatus );
        if ( !(nRetCode && !pIStatus) ) return 0;

        initialized = true;

        /*
        plugin_name.set_string_ansi ( pIAC->GetName() );
        if ( pIEE->GetCount() > 0 ) {
            string8 tmp;
            tmp.set_string_ansi ( pIEE->GetExt(0) );
            plugin_ext.set_string_lower ( tmp );
        }
        */

        return 1;
    }

    int init_decoder()
    {
        int nRetCode = pSrc->Open ( path, &pIU );
        if ( !(nRetCode && pIU) ) return 0;

        nRetCode = pIU->QueryInterface ( IID_IAudioRawReader, (void**)&pIAURR );
        if ( !(!nRetCode && pIAURR) ) return 0;

        format = pIAURR->GetRawFormat();

        nRetCode = pIU->QueryInterface ( IID_IProcess, (void**)&pIProcess );
        if ( !(!nRetCode && pIProcess) ) return 0;

        nRetCode = pIProcess->Start ( &pIStatus );
        if ( !(nRetCode && !pIStatus) ) return 0;

        nRetCode = pIAURR->QueryInterface ( IID_IExAudioInfo, (void**)&pIEAI );
        if ( !(!nRetCode && pIEAI) ) pIEAI = 0; // return 0;

        nRetCode = pIAURR->QueryInterface ( IID_ISeekable, (void**)&pIS );
        if ( !(!nRetCode && pIS) ) pIS = 0; // return 0;

        nRetCode = pIAURR->QueryInterface ( IID_IInfoReader, (void**)&pIR );
        if ( !(!nRetCode && pIR) ) pIR = 0; // return 0;

        nRetCode = pIAURR->QueryInterface ( IID_IAudioSource, (void**)&pIAS );
        if ( !(!nRetCode && pIAS) ) pIAS = 0; // return 0;

        _bitrate = 0;
        if ( pIEAI ) _bitrate = pIEAI->GetBitrate();

        _duration = 0;
        if ( pIAS ) _duration = pIAS->GetDuration();

        return 1;
    }

private:
    char path[_MAX_PATH];
    char cwd[_MAX_PATH];
    char plugin_dir[_MAX_PATH];
    //unsigned char *buffer;
    unsigned int _bitrate;
    ULONGLONG _duration;
    bool initialized;

    SWavFormat format;

    CAudioPluginMgr pluginMgr;
    IStatus *pIStatus;

    HINSTANCE hDLL;
    LPNERO_PLUGIN_GetPrimaryAudioObject GetPrimaryAudioObject;
    LPNERO_PLUGIN_ReadyToFinish ReadyToFinish;
    IAudioComponent *pIA;

    IComponentEnum *pICU;
    IAudioComponent *pIAC;
    IExtEnum *pIEE;
    IURLAudioSourceFactory *pSrc;
    IUnknown *pIU;
    IExAudioInfo *pIEAI;
    ISeekable *pIS;
    IInfoReader *pIR;
    IAudioSource *pIAS;

    IAudioRawReader *pIAURR;
    IProcess *pIProcess;
};

// -------------------------------------

typedef struct {
    nero_dec *decoder;

    unsigned char sample_buffer[CHUNKSIZE];
    int bpos, bwaiting;

    unsigned int srate, nch, bps;
    double seconds, bitrate;

    DWORD Options;
} INPUT;

typedef struct {
    nero_enc *encoder;

    unsigned int nch, bps;
    unsigned int required_srate;
    DWORD Options;
} OUTPUT;

// Options ------

__declspec (dllexport) DWORD FAR PASCAL FilterGetOptions ( HWND hWnd, HINSTANCE hInst, long lSamprate, WORD wChannels, WORD wBitsPerSample, DWORD dwOptions )
{
    nero_enc *enc = new nero_enc();
    if ( enc ) {
        enc->config();
        delete enc;
    }

    return 0;
}

__declspec (dllexport) DWORD FAR PASCAL FilterOptions ( HANDLE hInput )
{
    if ( !hInput ) return 0;

    DWORD options = 0;
    INPUT far *i = (INPUT far *)GlobalLock ( hInput );
    options = i->Options;
    GlobalUnlock ( hInput );

    return options;
}

__declspec (dllexport) DWORD FAR PASCAL FilterOptionsString ( HANDLE hInput, LPSTR szString )
{
    if ( !hInput || szString == NULL ) return 0;

    DWORD options = 0;
    INPUT far *i = (INPUT far *)GlobalLock ( hInput );

    if ( i->decoder ) {
        wsprintf ( szString, "MPEG-4 Audio\nBitrate: %u kbps", (unsigned int)i->bitrate );
        options = i->Options;
    } else {
        szString[0] = '\0';
        options = 0;
    }

    GlobalUnlock ( hInput );

    return options;
}

// End Options ---
// General ------

__declspec (dllexport) short FAR PASCAL QueryCoolFilter ( COOLQUERY far *cq )
{
    strcpy ( cq->szName,      "Nero MP4" );
    strcpy ( cq->szCopyright, "Nero Digital Audio (MPEG-4 AAC)" );
    strcpy ( cq->szExt,       "MP4" );
    strcpy ( cq->szExt2,      "M4A" );
    strcpy ( cq->szExt3,      "" );
    strcpy ( cq->szExt4,      "" );

    cq->lChunkSize = CHUNKSIZE;
    cq->dwFlags    = QF_CANLOAD | QF_CANSAVE | QF_HASOPTIONSBOX | QF_RATEADJUSTABLE | QF_NOASKFORCONVERT;
    cq->Stereo8    = 0xFF;
    cq->Stereo16   = 0xFF;
    cq->Stereo24   = 0xFF;
    cq->Stereo32   = 0xFF;
    cq->Mono8      = 0xFF;
    cq->Mono16     = 0xFF;
    cq->Mono24     = 0xFF;
    cq->Mono32     = 0xFF;
    cq->Quad32     = 0xFF;
    cq->Quad16     = 0xFF;
    cq->Quad8      = 0xFF;

    return C_VALIDLIBRARY;
}

__declspec (dllexport) BOOL FAR PASCAL FilterUnderstandsFormat ( LPSTR lpstrFilename )
{
    if ( !lpstrFilename || !lpstrFilename[0] ) return FALSE;

    const char *ext = strrchr ( lpstrFilename, '.' );
    if ( !ext ) return FALSE;
    return ( !stricmp (ext, ".mp4") || !stricmp (ext, ".m4a") );
}

__declspec (dllexport) VOID FAR PASCAL GetSuggestedSampleType ( LONG *lplSamprate, WORD *lplwBitsPerSample, WORD *lpwChannels )
{
    int sr  = *lplSamprate;
    int ch  = *lpwChannels;
    int bps = *lplwBitsPerSample;

    *lplSamprate = 0;
    *lpwChannels = 0;
    *lplwBitsPerSample = BPS;

    if ( ch > CN_MAX_CHANNELS ) *lpwChannels = CN_MAX_CHANNELS;
    if ( required_sample_rate > 0 && sr != required_sample_rate ) *lplSamprate = required_sample_rate;
}

__declspec (dllexport) long FAR PASCAL FilterGetFileSize ( HANDLE hInput )
{
    long size_in_bytes = 0;

    if ( hInput ) {
        INPUT FAR *i = (INPUT FAR *)GlobalLock ( hInput );
        if ( i->decoder ) {
            __int64 samples = (__int64)ceil(i->seconds * i->srate);
            size_in_bytes = (long)(samples * i->nch * (i->bps/8));
        }
        GlobalUnlock ( hInput );
    }

    return size_in_bytes;
}

// End General ---
// Reading ------

__declspec (dllexport) HANDLE FAR PASCAL OpenFilterInput ( LPSTR lpstrFilename, long far *lSamprate, WORD far *wBitsPerSample, WORD far *wChannels, HWND hWnd, long far *lChunkSize )
{
    HANDLE hInput;

    hInput = GlobalAlloc ( GMEM_MOVEABLE | GMEM_SHARE | GMEM_ZEROINIT, sizeof (INPUT) );
    if ( hInput ) {
        INPUT far *i = (INPUT far *)GlobalLock ( hInput );

        i->decoder = new nero_dec();
        if ( !i->decoder ) {
            GlobalUnlock ( hInput );
            GlobalFree ( hInput );
            return NULL;
        }

        if ( !i->decoder->init (lpstrFilename, &i->srate, &i->nch, &i->bps, &i->seconds, &i->bitrate) ) {
            delete i->decoder;
            GlobalUnlock ( hInput );
            GlobalFree ( hInput );
            return NULL;
        }

        if ( i->nch > CN_MAX_CHANNELS ) {
            char temp[256];
            wsprintf ( temp, "Maximum of %i channels supported, file has %i.", CN_MAX_CHANNELS, (int)i->nch );
            infobox ( temp );
            delete i->decoder;
            GlobalUnlock ( hInput );
            GlobalFree ( hInput );
            return NULL;
        }

        *lSamprate = i->srate;
        *wBitsPerSample = i->bps;
        *wChannels = i->nch;
        *lChunkSize = (*wChannels) * (*wBitsPerSample)/8 * 2048;

        GlobalUnlock ( hInput );
    }

    return hInput;
}

__declspec (dllexport) void FAR PASCAL CloseFilterInput ( HANDLE hInput )
{
    if ( hInput ) {
        INPUT far *i = (INPUT far *)GlobalLock ( hInput );

        if ( i->decoder ) delete i->decoder;

        GlobalUnlock ( hInput );
        GlobalFree ( hInput );
    }
}

__declspec (dllexport) DWORD FAR PASCAL ReadFilterInput ( HANDLE hInput, unsigned char far *buffer, long lBytes )
{
    unsigned char far *buf = buffer;
    unsigned int read = 0;

    if (hInput) {
        INPUT far *i = (INPUT far *)GlobalLock ( hInput );
        unsigned int b;

        if ( i->bwaiting > 0 ) {
            b = ( i->bwaiting > lBytes )  ?  lBytes  :  i->bwaiting;
            memcpy ( buf, i->sample_buffer + i->bpos, b );
            buf += b;
            i->bwaiting -= b;
            if ( i->bwaiting == 0 ) i->bpos = 0; else i->bpos += b;
            read += b;
        }

        while ( read < (unsigned)lBytes ) {
            unsigned int bytes, srate, bps, nch;

            if ( !i->decoder->decode (i->sample_buffer, &bytes, &srate, &bps, &nch) ) break;

            if ( read + bytes <= (unsigned)lBytes ) {
                b = bytes;
            } else {
                b = lBytes - read;
                i->bpos = b;
                i->bwaiting = bytes - i->bpos;
            }

            memcpy ( buf, i->sample_buffer, b );
            buf += b;
            read += b;
        }

        GlobalUnlock ( hInput );
    }

    return read;
}

// End Reading ---
// Writing ------

__declspec (dllexport) HANDLE FAR PASCAL OpenFilterOutput ( LPSTR lpstrFilename, long lSamprate, WORD wBitsPerSample, WORD wChannels, long lSize, long far *lpChunkSize, DWORD dwOptions )
{
    HANDLE hOutput;

    hOutput = GlobalAlloc ( GMEM_MOVEABLE | GMEM_SHARE | GMEM_ZEROINIT, sizeof (OUTPUT) );
    if ( hOutput ) {
        OUTPUT far *o = (OUTPUT far *)GlobalLock ( hOutput );

        o->nch = wChannels;
        o->bps = wBitsPerSample;
        o->Options = dwOptions;

        o->encoder = new nero_enc();
        if ( !o->encoder ) {
            GlobalUnlock ( hOutput );
            GlobalFree ( hOutput );
            infobox ( "Failed to create new encoder." );
            return NULL;
        }

        if ( !o->encoder->init(lpstrFilename, lSamprate, o->nch, o->bps, &o->required_srate) ) {
            delete o->encoder;
            GlobalUnlock ( hOutput );
            GlobalFree ( hOutput );
            return NULL;
        }

        if ( (o->bps != BPS) || ((unsigned)lSamprate != o->required_srate) ) {
            required_sample_rate = o->required_srate;
            delete o->encoder;
            GlobalUnlock ( hOutput );
            GlobalFree ( hOutput );
            return NULL;
        }

        *lpChunkSize = CHUNKSIZE;

        GlobalUnlock ( hOutput );
    }

    return hOutput;
}

__declspec (dllexport) void FAR PASCAL CloseFilterOutput ( HANDLE hOutput )
{
    if ( hOutput ) {
        OUTPUT far *o = (OUTPUT far *)GlobalLock ( hOutput );

        if ( o->encoder ) delete o->encoder;

        GlobalUnlock ( hOutput );
        GlobalFree ( hOutput );
    }

    required_sample_rate = 0;
}

__declspec (dllexport) DWORD FAR PASCAL WriteFilterOutput ( HANDLE hOutput, unsigned char far *buffer, long lBytes )
{
    if ( hOutput ) {
        OUTPUT far *o = (OUTPUT far *)GlobalLock ( hOutput );

        unsigned int samples = lBytes / (o->nch * o->bps/8);

        if ( !o->encoder->encode (buffer, samples, o->bps, o->nch) ) {
            GlobalUnlock ( hOutput );
            return 0;
        }
    }

    GlobalUnlock ( hOutput );

    return lBytes;
}

// End Writing ---
// DllMain ------

BOOL WINAPI DllMain ( HANDLE hModule, DWORD fdwReason, LPVOID lpReserved )
{
   switch ( fdwReason ) {
      case DLL_PROCESS_ATTACH:
        /* Code from LibMain inserted here.  Return TRUE to keep the
            DLL loaded or return FALSE to fail loading the DLL.
 
            You may have to modify the code in your original LibMain to
            account for the fact that it may be called more than once.
            You will get one DLL_PROCESS_ATTACH for each process that
            loads the DLL. This is different from LibMain which gets
            called only once when the DLL is loaded. The only time this
            is critical is when you are using shared data sections.
            If you are using shared data sections for statically
            allocated data, you will need to be careful to initialize it
            only once. Check your code carefully.
 
            Certain one-time initializations may now need to be done for
            each process that attaches. You may also not need code from
            your original LibMain because the operating system may now
            be doing it for you.
         */
         break;
 
      case DLL_THREAD_ATTACH:
         /* Called each time a thread is created in a process that has
            already loaded (attached to) this DLL. Does not get called
            for each thread that exists in the process before it loaded
            the DLL.
 
            Do thread-specific initialization here.
         */
         break;
 
      case DLL_THREAD_DETACH:
         /* Same as above, but called when a thread in the process
            exits.
 
            Do thread-specific cleanup here.
         */
         break;
 
      case DLL_PROCESS_DETACH:
         /* Code from _WEP inserted here.  This code may (like the
            LibMain) not be necessary.  Check to make certain that the
            operating system is not doing it for you.
         */
         break;
   }
 
   /* The return value is only used for DLL_PROCESS_ATTACH; all other
      conditions are ignored.  */
   return TRUE;   // successful DLL_PROCESS_ATTACH
}

// End DllMain ---
